#!/usr/bin/env python

"""
Specific classes for storing document information.

Copyright (C) 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.files import *
from bisect import bisect_right  # to find terms in the dictionary index

class FieldWriter(FileWriter):

    "Writing field data to files."

    def reset(self):
        self.last_docnum = 0

    def write_fields(self, docnum, fields):

        """
        Write for the given 'docnum', a list of 'fields' (integer, string pairs
        representing field identifiers and values respectively).
        """

        # Write the document number delta.

        self.write_number(docnum - self.last_docnum)

        # Write the number of fields.

        self.write_number(len(fields))

        # Write the fields themselves.

        for i, field in fields:
            self.write_number(i)
            self.write_string(field, 1) # compress

        self.last_docnum = docnum

class FieldReader(FileReader):

    "Reading field data from files."

    def reset(self):
        self.last_docnum = 0

    def read_fields(self):

        """
        Read fields from the file, returning a tuple containing the document
        number and a list of field (identifier, value) pairs.
        """

        # Read the document number.

        self.last_docnum += self.read_number()

        # Read the number of fields.

        nfields = self.read_number()

        # Collect the fields.

        fields = []
        i = 0

        while i < nfields:
            identifier = self.read_number()
            value = self.read_string(1) # decompress
            fields.append((identifier, value))
            i += 1

        return self.last_docnum, fields

    def read_document_fields(self, docnum, offset):

        """
        Read fields for 'docnum' at the given 'offset'. This permits the
        retrieval of details for the specified document, as well as scanning for
        later documents.
        """

        self.f.seek(offset)
        bad_docnum, fields = self.read_fields()
        self.last_docnum = docnum
        return docnum, fields

class FieldIndexWriter(FileWriter):

    "Writing field index details to files."

    def reset(self):
        self.last_docnum = 0
        self.last_offset = 0

    def write_document(self, docnum, offset):

        """
        Write for the given 'docnum', the 'offset' at which the fields for the
        document are stored in the fields file.
        """

        # Write the document number and offset deltas.

        self.write_number(docnum - self.last_docnum)
        self.write_number(offset - self.last_offset)

        self.last_docnum = docnum
        self.last_offset = offset

class FieldIndexReader(FileReader):

    "Reading field index details from files."

    def reset(self):
        self.last_docnum = 0
        self.last_offset = 0

    def read_document(self):

        "Read a document number and field file offset."

        # Read the document number delta and offset.

        self.last_docnum += self.read_number()
        self.last_offset += self.read_number()

        return self.last_docnum, self.last_offset

class FieldDictionaryWriter:

    "Writing field dictionary details."

    def __init__(self, field_writer, field_index_writer, interval):
        self.field_writer = field_writer
        self.field_index_writer = field_index_writer
        self.interval = interval
        self.entry = 0

    def write_fields(self, docnum, fields):

        "Write details of the document with the given 'docnum' and 'fields'."

        if self.entry % self.interval == 0:
            offset = self.field_writer.f.tell()
            self.field_writer.write_fields(docnum, fields)
            self.field_index_writer.write_document(docnum, offset)
        else:
            self.field_writer.write_fields(docnum, fields)

        self.entry += 1

    def close(self):
        self.field_writer.close()
        self.field_index_writer.close()

class FieldDictionaryReader:

    "Reading field dictionary details."

    def __init__(self, field_reader, field_index_reader):
        self.field_reader = field_reader
        self.field_index_reader = field_index_reader

        self.docs = []
        try:
            while 1:
                self.docs.append(self.field_index_reader.read_document())
        except EOFError:
            pass

        # Large numbers for ordering purposes.

        if self.docs:
            self.max_offset = self.docs[-1][1]
        else:
            self.max_offset = None

    # Iterator convenience methods.

    def __iter__(self):
        self.rewind()
        return self

    def next(self):
        try:
            return self.read_fields()
        except EOFError:
            raise StopIteration

    # Sequential access methods.

    def rewind(self):
        self.field_reader.rewind()

    def read_fields(self):

        "Return the next document number and fields."

        return self.field_reader.read_fields()

    # Random access methods.

    def get_fields(self, docnum):

        "Read the fields of the document with the given 'docnum'."

        i = bisect_right(self.docs, (docnum, self.max_offset)) - 1

        # Get the entry position providing the term or one preceding it.

        if i == -1:
            return None

        found_docnum, offset = self.docs[i]

        # Read from the fields file.

        found_docnum, fields = self.field_reader.read_document_fields(found_docnum, offset)

        # Scan for the document, if necessary.

        try:
            while docnum > found_docnum:
                found_docnum, fields = self.field_reader.read_fields()
        except EOFError:
            pass

        # If the document is found, return the fields.

        if docnum == found_docnum:
            return fields
        else:
            return None

    def close(self):
        self.field_reader.close()
        self.field_index_reader.close()

# vim: tabstop=4 expandtab shiftwidth=4
